'''
avlib 0.4.6

(c) 2003 Jan ONDREJ (SAL) <ondrejj@salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

import os,sys,re,string,time,random,traceback,signal,resource,struct,socket

BUFSIZE=10240
ROOT_PATH=''
MASTER_PID=os.getpid()
allowed_chars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789._-'

class ScannerError(Exception):
      """Any Error Of Any Scanner"""

class debug_class:
  '''
  Debug class. Used for debugging informations.
  '''
  debug_level=2
  def __init__(self):
      self.set_logfile('-')
  def set_logfile(self,logfile):
      if logfile!='-':
        self.fd=os.open(logfile,
          os.O_CREAT|os.O_WRONLY|os.O_APPEND,0640)
        self.logfd=os.fdopen(self.fd,'a',0)
        #os.chmod(logfile,0640)
      else:
        self.logfd=os.fdopen(os.dup(1),'w',0)
        self.fd=1
      sys.stdout=self.logfd
      sys.stderr=self.logfd
  def set_level(self,level):
      self.debug_level=level
  def echo(self,level,*s):
      if self.debug_level>=level:
        o=''
        for i in s:
          if type(i)==type([]):
            for x in i:
              o=o+str(x)
          else:
            o=o+str(i)
        p=os.getpid()
        while 1:
          try:
            self.logfd.write("%5s: %s\n" % (p,o))
            break
          except IOError,(ec,es):
            if ec!=4:
              break
  def traceback(self,level,arg):
      self.echo(level,arg+str(traceback.format_exception(sys.exc_type,
                       sys.exc_value, sys.exc_traceback)))
  def traceback_value_str(self):
      return string.join(traceback.format_exception_only(
                             sys.exc_type, sys.exc_value),', ')

debug=debug_class()

def set_root_path(np):
    global ROOT_PATH
    ROOT_PATH=np

def normalize_filename(sfn):
    '''replace !allowed_chars from filename with _'''
    ofn=''
    for chp in range(len(sfn)):
      if not sfn[chp] in allowed_chars:
        ofn=ofn+'_'
      else:
        ofn=ofn+sfn[chp]
    return ofn

class popen:
  '''
  POpen implementation (with some changes)
  '''
  def __init__(self, cmd, wd='', resource_limits={}):
      self.exitstatus,self.killsignal,self.execerror,self.pid=-1,-1,'',0
      self.pcr,self.pcw = os.pipe()
      self.cpr,self.cpw = os.pipe()
      self.oldsigchld=signal.getsignal(signal.SIGCHLD)
      signal.signal(signal.SIGUSR1,self.sighandler)
      signal.signal(signal.SIGCHLD,self.sighandler)
      self.pid = os.fork()
      if self.pid == 0: # child
        signal.signal(signal.SIGUSR1,signal.SIG_IGN)
        signal.signal(signal.SIGCHLD,signal.SIG_IGN)
        os.close(self.pcw)
        os.close(self.cpr)
        os.dup2(self.pcr,0); os.close(self.pcr)
        os.dup2(self.cpw,1)
        os.dup2(self.cpw,2); os.close(self.cpw)
        try:
          if wd:
            os.chdir(wd)
          for key,value in resource_limits.items():
            resource.setrlimit(key,(value,value))
          debug.echo(4,"POpen: Running: ",str(cmd))
          os.execvp(cmd[0],cmd)
        except os.error,(ec,es):
          debug.echo(3,"POpen: os.error: ",es+str([cmd[0]]))
          os.write(1,es+str([cmd[0]]))
        os._exit(10)
      os.close(self.pcr)
      os.close(self.cpw)
      self.tocmd=os.fdopen(self.pcw, 'w', 0)
      self.fromcmd=os.fdopen(self.cpr, 'r', 0)
  def sighandler(self,sn,stack):
      if sn==signal.SIGUSR1:
        self.execerror=os.read(self.cpr,1024)
      else:
        self.wait()
        debug.echo(4,"POpen SIG=",sn,", pid=",self.pid,
              ", exitstatus=",self.exitstatus,
              ", signal=",self.killsignal)
  def wait(self):
      try:
        pid,es=os.waitpid(self.pid,0)
        self.exitstatus=es/256
        self.killsignal=es%256
      except OSError:
        if self.exitstatus<0:
          self.exitstatus=0 # pid already exited, but no exit status
      signal.signal(signal.SIGCHLD,self.oldsigchld)
      return self.exitstatus
  def readlines(self):
      try:
        return self.fromcmd.readlines()
      except IOError,(ec,es):
        if ec==4: #Interrupted system call
          return []
        else:
          raise
  def write(self,s):
      if self.exitstatus<0:
        return self.tocmd.write(s)
      else:
        return 0
  def close(self,rw=3):
      if rw&1:
        self.tocmd.flush()
        os.close(self.pcw)
      if rw&2:
        os.close(self.cpr)

def rlistdir(rootpath,path=''):
    '''
    list files and directories recursively
    
    return values: [filelist],size
      [filelist] is an array of file and directory names
      size is added filesizes
    '''
    files,fsize=[],0L
    try:
      listdir=os.listdir(os.path.join(rootpath,path))
    except:
      return [],0L
    for d in listdir:
      stat=os.lstat(os.path.join(rootpath,path,d))
      if stat[0] & 0x4000: # is_dir
        rld=rlistdir(rootpath,os.path.join(path,d))
        files=files+rld[0]
        fsize=fsize+rld[1]
      else: # is_file or is_symlink or ...
        fsize=fsize+stat[6]
        files.append(os.path.join(path,d))
    return files,fsize

def randomchars(chars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'):
    '''
    Return 10 random characters. Characters are randomly selected
    from chars array.
    '''
    s=''
    r=random.Random()
    for i in range(10):
      s=s+chars[int(r.random()*len(chars))]
    return s

class mktemp:
      '''
      make a temporary file or directory
      Makeing of file:
        mktemp('filename-prefix','.suffix','w',0600).autorm()
      Makeinf od directory:
        mktemp('filename-prefix','.suffix','d',0600).autorm()
      '''
      def __init__(self,prefix,suffix='',flags='w',umode=-1):
          self.flag_autorm=0
          self.flags=flags
          while 1:
            s=randomchars()
            self.name=prefix+s+suffix
            self.name_no_prefix=s+suffix
            try:
              if flags=='d':
                if umode<0:
                  umode=0700
                os.mkdir(self.name)
                os.chmod(self.name,umode)
                if (os.stat(self.name)[0] & 0170000)==040000:
                  return
              else:
                if umode<0:
                  umode=0600
                self.fd=os.open(self.name,os.O_WRONLY|os.O_CREAT|os.O_EXCL,umode)
                self.f=os.fdopen(self.fd,'w',BUFSIZE)
                if (os.stat(self.name)[0] & 0170000)==0100000:
                  return
            except OSError,(ec,es):
              if not (ec in [17,21]):
                raise
      def __del__(self):
          if self.flag_autorm:
            self.rm()
      def rm(self):
          if self.flags=='d':
            os.rmdir(self.name)
          else:
            try:
              self.f.close()
            except:
              pass
            os.unlink(self.name)
      def autorm(self):
          self.flag_autorm=1
          return self

def tempcleanup(filename,s=''):
    '''
    Safe cleanup for a file.
    '''
    try:
      os.unlink(filename)
    except OSError,(er,es):
      debug.echo(0,s+": ERROR: OSError: ",es,filename)

def rdel(paths):
    '''
    deletes directory recursively
      paths is an array of directories or files
    '''
    for dirname in paths:
      try:
        if os.lstat(dirname)[0] & 0x4000: # is_dir
          for fname in os.listdir(dirname):
            rdel([os.path.join(dirname,fname)])
          os.rmdir(dirname)
        else: # is fname or symlink
          os.unlink(dirname)
      except OSError,(ec,es):
        if not ec in [2]:
          debug.echo(3,"rdel: ",debug.traceback_value_str())

def fromhdr():
    '''
    Return a "From local@... DATE" string. String is by RFC.
    '''
    return "From local@local.local"+\
      time.strftime("  %c\r\n",time.localtime(time.time()))

def socket_settimeout(sock,seconds):
    '''
    Set a timeout for a socket. Timeout is in seconds.
    '''
    sock.setsockopt(socket.SOL_SOCKET,socket.SO_RCVTIMEO,
      struct.pack('ll',seconds,0))

class ascanner:
  '''
  Default scanner user for building all other realscanners.
  '''
  name='AScanner'
  is_spamscan=0
  is_multistream=0
  filename='stream'
  xhdr=[]
  scanner=0
  #def __init__(self,arg):
  #    self.arg=arg
  def destroy(self):
      pass
  def scanstream(self,stream):
      raise ScannerError,'Not implemented'
  def scanfile(self,files,dirname=''):
      raise ScannerError,'Not implemented'
  def reinit(self):
      pass
  def xheader(self):
      return self.xhdr

class interscanner(ascanner):
  '''
  Default scanner used for building all other interscanners.
  '''
  name='InterScanner'
  scanners=[]
  #def __init__(self,scanner):
  #    self.scanner=scanner
  def rename(self,scanners):
      n=[]
      for s in scanners:
        n.append(s.name)
      self.name=self.name.replace("()","("+string.join(n,',')+")")
      if self.name=="":
        self.name=string.join(n,',')
  def destroy(self):
      self.scanner.destroy()
  def reinit(self):
      self.scanner.reinit()
  def xheader(self):
      return self.scanner.xheader()
